package ch.unibe.piccola.compiler;

import junit.framework.TestCase;
import junit.framework.TestSuite;
import ch.unibe.piccola.code.Code;
import ch.unibe.piccola.code.CodeFactory;
import ch.unibe.piccola.parser.Parser;
import ch.unibe.piccola.parser.ParserException;
import ch.unibe.piccola.parser.ParserFactory;
import ch.unibe.piccola.source.StringSourceCode;
import ch.unibe.piccola.syntax.Node;
import ch.unibe.piccola.syntax.Syntax;

/**
 * @author   Stefan Kneubuehl
 * @version  3.4c (Nov 14, 2001)
**/
public class CodeMakerTest extends TestCase {

    // ------------------------------------------------------------------------
    // private static
    // ------------------------------------------------------------------------

    private static final String LABEL_ARG = Syntax.labelArg();

    private static final String LABEL_DYNAMIC = Syntax.labelDynamic();

    // ------------------------------------------------------------------------
    // test suite
    // ------------------------------------------------------------------------

    public CodeMakerTest(String name)
    {
        super(name);
    }

    public static TestSuite suite()
    {
        return new TestSuite(CodeMakerTest.class);
    }

    public static void main(String[] args)
    {
        junit.textui.TestRunner.run(suite());
    }

    // ------------------------------------------------------------------------
    // test environment
    // ------------------------------------------------------------------------

    public void setUp()
    {
        compiler = new DefaultCompiler();
        parser = ParserFactory.INSTANCE.newJavaCCParser();
        cf = new CodeFactory();
    }

    private DefaultCompiler compiler;

    private CodeFactory cf;

    private Parser parser;

    private Code compile(Node syntax)
    {
        try {
            return compiler.compile(syntax);
        }
        catch (Throwable ex) {
            fail(syntax + " yields " + ex);
        }
        return null;
    }

    private Node parse(String code)
    {
        try {
            return parser.parse(new StringSourceCode(code));
        }
        catch (ParserException ex) {
            fail("parsing failed: while parsing \"" + code + "\" " + ex);
        }
        catch (Throwable ex) {
            fail("Unexpected exception: " + ex);
        }
        return null;
    }

    private Code compile(String code)
    {
        return compile(parse(code));
    }

    private void assertCode(String s)
    {
        cf.emitExit(null);
        Code f = compile(s);
        assertEquals(cf.code(), f);
    }

    // ------------------------------------------------------------------------
    // tests
    // ------------------------------------------------------------------------

    public void testAbstraction1()
    {
        cf.emitRoot(null);
        cf.emitLookup(null, LABEL_DYNAMIC);
        cf.emitProject(null, LABEL_ARG);
        cf.emitBind(null, "param");
        cf.emitExtend(null);
        cf.emitPushContext(null);
        cf.emitEmpty(null);
        cf.emitPopContext(null);
        cf.emitExit(null);
        Code service = cf.code();
        cf.clear();
        cf.emitService(null, service);
        assertCode("\\param: ()");
    }

    public void testAbstraction2()
    {
        cf.emitEmpty(null);
        cf.emitExit(null);
        Code service = cf.code();
        cf.clear();
        cf.emitService(null, service);
        assertCode("\\: ()");
    }

    public void testApplication()
    {
        cf.emitLookup(null, "a");
        cf.emitLookup(null, LABEL_DYNAMIC);
        cf.emitLookup(null, "b");
        cf.emitBind(null, LABEL_ARG);
        cf.emitExtend(null);
        cf.emitApply(null);
        assertCode("a b");
    }

    public void testBinding()
    {
        cf.emitEmpty(null);
        cf.emitBind(null, "x");
        assertCode("x = ()");
    }

    public void testEmpty()
    {
        cf.emitEmpty(null);
        assertCode("()");
    }

    public void testExtension1()
    {
        cf.emitEmpty(null);
        cf.emitExit(null);
        Code service = cf.code();
        cf.clear();
        cf.emitEmpty(null);
        cf.emitBind(null, "a");
        cf.emitService(null, service);
        cf.emitExtend(null);
        assertCode("(a = ()), \\: ()");
    }

    public void testExtension2()
    {
        cf.emitEmpty(null);
        cf.emitExit(null);
        Code service = cf.code();
        cf.clear();
        cf.emitService(null, service);
        cf.emitBind(null, "a");
        cf.emitService(null, service);
        cf.emitBind(null, "b");
        cf.emitExtend(null);
        cf.emitBind(null, "Hook");
        cf.emitLookup(null, "cont");
        cf.emitExtend(null);
        assertCode("(Hook = ((a: ()), (b: ()))), cont");
    }

// TODO: doesnt work due to temporal change of fixed point lookup
//    public void testFixedPoint()
//    {
//        cf.emitRoot(null);
//        cf.emitLookup(null, "x");
//        cf.emitLookup(null, S_DYNAMIC);
//        cf.emitEmpty(null);
//        cf.emitBind(null, S_ARG);
//        cf.emitExtend(null);
//        cf.emitApply(null);
//        cf.emitProject(null, "x");
//        cf.emitBind(null, "x");
//        cf.emitExtend(null);
//        cf.emitPushContext(null);
//        // service code
//        cf.emitLookup(null, "x");
//        // service code end
//        cf.emitPopContext(null);
//        cf.emitExit(null);
//        Code service = cf.code();
//
//        cf.clear();
//        cf.emitRoot(null);
//        cf.emitEmpty(null);
//        cf.emitVariable(null);
//        cf.emitBind(null, "x");
//        cf.emitExtend(null);
//        cf.emitPushContext(null);
//
//        cf.emitLookup(null, "x");
//        cf.emitProject(null, S_SET);
//        cf.emitLookup(null, S_DYNAMIC);
//        cf.emitService(null, service);
//        cf.emitBind(null, "f");
//        cf.emitBind(null, "x");
//        cf.emitBind(null, S_ARG);
//        cf.emitExtend(null);
//
//        cf.emitApply(null);
//        cf.emitProject(null, "x");
//        cf.emitPopContext(null);
//        cf.emitBind(null, "x");
//        assertCode("def x = f: x");
//    }

    public void testLabel()
    {
        cf.emitLookup(null, "label");
        assertCode("label");
    }

    public void testConst1()
    {
        cf.emitJavaObject(null, new String("HELLO"));
        assertCode("\"HELLO\"");
    }

    public void testConst2()
    {
        cf.emitJavaObject(null, new Integer(1234));
        assertCode("1234");
    }

    public void testProjection()
    {
        cf.emitLookup(null, "x");
        cf.emitProject(null, "label");
        assertCode("x.label");
    }

    public void testRoot()
    {
        cf.emitRoot(null);
        assertCode("root");
    }

    public void testSandbox1()
    {
        cf.emitEmpty(null);
        cf.emitBind(null, "a");
        cf.emitPopValue(null);
        cf.emitEmpty(null);
        cf.emitLookup(null, "a");
        cf.emitExtend(null);
        assertCode("(root = (a = ())), a");
    }

    public void testSandbox2()
    {
        cf.emitEmpty(null);
        cf.emitBind(null, "a");
        cf.emitPushContext(null);
        cf.emitLookup(null, "a");
        cf.emitPopContext(null);
        assertCode("root = (a = ()), a");
    }

}

