package ch.unibe.piccola.compiler;

import ch.unibe.piccola.syntax.*;
import ch.unibe.piccola.source.SourceRef;

/**
 * Implements the lookup[.] transformation as defined in the Piccola Bible
 * page 82ff.
 *
 * @author   Stefan Kneubuehl
 * @version  3.4c (Dec 09, 2002)
**/
public final class LookupTransformation extends SimplifiedNodeVisitor
{

    // ------------------------------------------------------------------------
    // public static
    // ------------------------------------------------------------------------

    public static final Node apply(String label, Node node)
    {
        return new LookupTransformation(label).visit(node);
    }

    // ------------------------------------------------------------------------
    // private static
    // ------------------------------------------------------------------------

    private static final String R_LABEL = "R";

    // ------------------------------------------------------------------------
    // public
    // ------------------------------------------------------------------------

    /**
     * Applies the following transformations:
     *     lookup[x].(\x: A) = \x: A
     *     lookup[x].(\y: A) = \y: (R, x -> x.().x); A
     * See Piccola Bible page 83.
     *
     * Temporarily changed semantics in order to address fixedPointFunnies.
    **/
    public final Object visitAbstraction(Abstraction node)
    {
        if (m_label.equals(node.getParameter())) {
            return node;
        }
        else {
            return sf.newAbstraction(node.getSourceRef(),
                apply(m_label, node.getChild()), node.getParameter());
        }
//        if (m_label.equals(node.getParameter())) {
//            return node;
//        }
//        else {
//            SourceRef sr = node.getSourceRef();
//            Node F = sf.newBinding(sr, xCallX(sr), m_label);
//            F = sf.newExtension(sr, sf.newRoot(sr), F);
//            F = sf.newSandbox(sr, F, node.getChild());
//            return sf.newAbstraction(sr, F, node.getParameter());
//        }
    }

    /**
     * Applies the following transformations:
     *     lookup[x].(def x -> A) = def x -> A
     *     lookup[x].(def y -> A) = def y -> lookup[x].A
    **/
    public final Object visitFixedPoint(FixedPoint node)
    {
        if (node.getLabel().equals(m_label)) {
            return node;
        }
        else {
            return sf.newFixedPoint(
                node.getSourceRef(), apply(m_label, node.getChild()), node.getLabel()
            );
        }
    }

    /**
     * Applies the following transformations:
     *     lookup[x].y = y
     *     lookup[x].x = x(); x
     * See Piccola Bible page 82.
    **/
    public final Object visitLabel(Label node)
    {
        if (node.getLabel().equals(m_label)) {
            return xCallX(node.getSourceRef());
        }
        else {
            return node;
        }
    }

    /**
     * Applies the following transformation:
     *     lookup[x].R = R \ x
     * See Piccola Bible page 83.
    **/
    public final Object visitRoot(Root node)
    {
       return sf.newHide(node.getSourceRef(), node, m_label);
    }

    /**
     * Applies the following transformation:
     *     lookup[x].(A; B) = r -> (lookup[x].A), x -> x;
     *                        r, x -> d.x(), r;
     *                        lookup[x].B
     * See Piccola Bible p. 84.
    **/
    public final Object  visitSandbox(Sandbox node)
    {
        SourceRef sr = node.getSourceRef();
        Node A = visit(node.getLeftChild());
        Node B = node.getRightChild();
        if (null != B) {
            B = visit(B);
        }
        // r
        Node R = sf.newLabel(sr, R_LABEL);
        // r = lookup[x].A
        A = sf.newBinding(sr, A, R_LABEL);
        // x = x
        Node C1 = sf.newBinding(sr, sf.newLabel(sr, m_label), m_label);
        // (r = lookup[x].A, x = x)
        C1 = sf.newExtension(sr, A, C1);
        // x(), r
        Node C2 = sf.newExtension(sr, xCall(sr), R);
        // \: (x(), r)
        C2 = sf.newAbstraction(sr, C2, "");
        // x: (x(), r)
        C2 = sf.newBinding(sr, C2, m_label);
        // r, x: (x(), r)
        C2 = sf.newExtension(sr, R, C2);
        // root = ((r = lookup[x].A, x = x), root = (r, x: (x(), r)), lookup[x].B)
        return sf.newSandbox(sr, C1, sf.newSandbox(sr, C2, B));
    }

    // ------------------------------------------------------------------------
    // private
    // ------------------------------------------------------------------------

    private LookupTransformation(String label)
    {
        m_label = label;
    }

    private final String m_label;

    /**
     * x()
    **/
    private final Node xCall(SourceRef sr)
    {
        Node node = sf.newLabel(sr, m_label);
        return sf.newApplication(sr, node, sf.newEmpty(sr));
    }

    /**
     * x().x
    **/
    private final Node xCallX(SourceRef sr)
    {
        return sf.newProjection(sr, xCall(sr), m_label);
    }

}