package ch.unibe.piccola.compiler;

import junit.framework.TestCase;
import junit.framework.TestSuite;
import ch.unibe.piccola.parser.Parser;
import ch.unibe.piccola.parser.ParserException;
import ch.unibe.piccola.parser.ParserFactory;
import ch.unibe.piccola.source.StringSourceCode;
import ch.unibe.piccola.syntax.Node;
import ch.unibe.piccola.syntax.SyntaxFactory;

/**
 * Tests for LookupTransformation.
 *
 * @author    Stefan Kneubuehl
 * @version   3.4c (Jul 19, 2002)
**/
public class LookupTransformationTest extends TestCase {

    // ------------------------------------------------------------------------
    // test suite
    // ------------------------------------------------------------------------

    public LookupTransformationTest(String name)
    {
        super(name);
    }

    public static TestSuite suite()
    {
        return new TestSuite(LookupTransformationTest.class);
    }

    public static void main(String[] args)
    {
        junit.textui.TestRunner.run(suite());
    }

    // ------------------------------------------------------------------------
    // test environment
    // ------------------------------------------------------------------------

    public void setUp()
    {
        parser = ParserFactory.INSTANCE.newJavaCCParser();
        sf = SyntaxFactory.INSTANCE;
    }

    private Parser parser;

    private SyntaxFactory sf;

    private Node parse(String code)
    {
        try {
            return parser.parse(new StringSourceCode(code));
        }
        catch (ParserException ex) {
            fail("parsing failed: while parsing \"" + code + "\" " + ex);
        }
        catch (Throwable ex) {
            fail("Unexpected exception: " + ex);
        }
        return null;
    }

    public Node lookupParse(String source)
    {
        return LookupTransformation.apply("x", siParse(source));
    }

    public Node siParse(String source)
    {
        return Simplifier.apply(parse(source));
    }

    public void assertLookup(String expect, String get)
    {
        assertEquals(siParse(expect), lookupParse(get));
    }

    // ------------------------------------------------------------------------
    // tests
    // ------------------------------------------------------------------------

// TODO: doesnt work due to temporal change of fixed point lookup
//    public void testFunny()
//    {
//        assertLookup("()", "((f = \\: \\: x), (g = f ()))");
//    }

    /**
     * Tests for abstractions. Asserts the following equalities:
     *     lookup[x].(\x: A) = \x: A
     *     lookup[x].(\y: A) = \y: (R, x -> x.().x); A
     * See Piccola Bible page 83.
    **/
// TODO: doesnt work due to temporal change of fixed point lookup
//    public void testAbstraction()
//    {
//        assertLookup("\\x: x", "\\x: x");
//        assertLookup("\\x: y", "\\x: y");
//        assertLookup("\\y: (root = (root, x = x().x), ())", "\\y: ()");
//        assertLookup("\\y: (root = (root, x = x().x), x)", "\\y: x");
//    }

    /**
     * Tests for applications. Asserts the following equality:
     *     lookup[x].(A B) = lookup[x].A lookup[x].B
    **/
    public void testApplication()
    {
        assertLookup("x().x y", "x y");
        assertLookup("y x().x", "y x");
    }

    /**
     * Tests for bindings. Asserts the following equalities:
     *     lookup[x].(x -> A) = x -> lookup[x].A
     *     lookup[x].(y -> A) = y -> lookup[x].A
    **/
    public void testBinding()
    {
        assertLookup("x = x().x", "x = x");
        assertLookup("y = x().x", "y = x");
    }

    /**
     * Test for collections. LookupTransformation does not support collections
     * and should throw an exception.
    **/
    public void testCollection()
    {
        try {
            LookupTransformation.apply("x", parse("{ a, b, c }"));
        }
        catch (Exception ex) { return; }
        fail("LookupTransformation should throw an exception.");
    }

    /**
     * Test for compound bindings. LookupTransformation does not support
     * compound bindings and should throw an exception.
    **/
    public void testCompoundBinding()
    {
        try {
            LookupTransformation.apply("x", parse("a.b = c"));
        }
        catch (Exception ex) { return; }
        fail("LookupTransformation should throw an exception.");
    }

    /**
     * Test for curried abstractions. LookupTransformation does not support
     * curried abstractions and should throw an exception.
    **/
    public void testCurriedAbstraction()
    {
        try {
            LookupTransformation.apply("x", parse("\\x y: ()"));
        }
        catch (Exception ex) { return; }
        fail("LookupTransformation should throw an exception.");
    }

    /**
     * Test for Empty syntax node.
    **/
    public void testEmpty()
    {
        assertLookup("()", "()");
    }

    /**
     * Tests for extensions. Asserts the following equality:
     *     lookup[x].(A, B) = x -> lookup[x].A, lookup[x].B
    **/
    public void testExtension()
    {
        assertLookup("x().x, y", "x, y");
        assertLookup("y, x().x", "y, x");
    }

    /**
     * Tests for nested fixed points. Asserts the following equalities:
     *     lookup[x].(def x -> A) = def x -> A
     *     lookup[x].(def y -> A) = def y -> lookup[x].A
    **/
    public void testFixedPoint()
    {
        assertLookup("def x = getx: x", "def x = getx: x");
        assertLookup("def y = x().x", "def y = x");
    }

    /**
     * Tests for hide. Asserts the following equalities:
     *     lookup[x].(A \ x) = lookup[x].A \ x
     *     lookup[x].(A \ y) = lookup[x].A \ y
    **/
    public void testHide()
    {
        Node e;
        Node f;

        // x \ y
        e = sf.newHide(null, siParse("x().x"), "y");
        f = sf.newHide(null, siParse("x"), "y");
        assertEquals(e, LookupTransformation.apply("x", f));

        // z \ y
        e = sf.newHide(null, siParse("z"), "y");
        assertEquals(e, LookupTransformation.apply("x", e));
    }

    /**
     * Test for infix applications. LookupTransformation does not support
     * infix applications and should throw an exception.
    **/
    public void testInfixApplication()
    {
        try {
            LookupTransformation.apply("x", parse("x + y"));
        }
        catch (Exception ex) { return; }
        fail("LookupTransformation should throw an exception.");
    }

    /**
     * Tests for variables. Asserts the following equalities:
     *     lookup[x].y = y
     *     lookup[x].x = x(); x
     * See Piccola Bible page 82.
    **/
    public void testLabel()
    {
        assertLookup("y", "y");
        assertLookup("x().x", "x");
    }

    /**
     * Test for Literal syntax node.
    **/
    public void testLiteral()
    {
        assertLookup("22", "22");
    }

    /**
     * Test for named abstractions. LookupTransformation does not support
     * named abstractions and should throw an exception.
    **/
    public void testNamedAbstraction()
    {
        try {
            LookupTransformation.apply("x", parse("f.g x y: ()"));
        }
        catch (Exception ex) { return; }
        fail("LookupTransformation should throw an exception.");
    }

    /**
     * Test for prefix applications. LookupTransformation does not support
     * prefix applications and should throw an exception.
    **/
    public void testPrefixApplication()
    {
        try {
            LookupTransformation.apply("x", parse("-x"));
        }
        catch (Exception ex) { return; }
        fail("LookupTransformation should throw an exception.");
    }

    /**
     * Tests for projections. Asserts the following equalities:
     *     lookup[x].(A.x) = (lookup[x].A).x
     *     lookup[x].(A.y) = (llokup[x].A).y
    **/
    public void testProjection()
    {
        assertLookup("(x().x).x", "x.x");
        assertLookup("y.x", "y.x");
        assertLookup("(x().x).y", "x.y");
    }

    /**
     * Test for Quote syntax node. LookupTransformation does not support this
     * syntax node type and should throw an exception.
    **/
    public void testQuote()
    {
        try {
            LookupTransformation.apply("x", parse("'x = 2"));
        }
        catch (Exception ex) { return; }
        fail("LookupTransformation should throw an exception.");
    }

    /**
     * Tests for root context. Asserts the following equality:
     *     lookup[x].R = R \ x
     * See Piccola Bible page 83.
    **/
    public void testRoot()
    {
        Node e = sf.newHide(null, sf.newRoot(null), "x");
        Node f = lookupParse("root");
        assertEquals(e, f);
    }

    /**
     * Tests for sandbox. Asserts the following equality:
     *     lookup[x].(A; B) = r -> (lookup[x].A), x -> x;
     *                        r, x -> \d.x(), r;
     *                        lookup[x].B
     * See Piccola Bible p. 84.
    **/
    public void testSandbox()
    {
        assertLookup("root = ((R = y), x = x), (root = (R, x: (x(), R)), z)",
                 "root = y, z");
        assertLookup("root = ((R = x().x), x = x), (root = (R, x: (x(), R)), z)",
                 "root = x, z");
        assertLookup("root = ((R = y), x = x), (root = (R, x: (x(), R)), x().x)",
                 "root = y, x");
        assertLookup("root = ((R = x().x), x = x), (root = (R, x: (x(), R)), x().x)",
                 "root = x, x");
    }

}