package ch.unibe.piccola.parser;

import java.util.ArrayList;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import ch.unibe.piccola.syntax.Node;
import ch.unibe.piccola.syntax.SyntaxFactory;
import ch.unibe.piccola.source.SourceRef;
import ch.unibe.piccola.source.StringSourceCode;

/**
 * @author   Stefan Kneubuehl
 * @version  3.4c (Oct 02, 2002)
**/
public class ParserTest extends TestCase
{

    // ------------------------------------------------------------------------
    // test suite
    // ------------------------------------------------------------------------

    public static void main(String[] args)
    {
        junit.textui.TestRunner.run(suite());
    }

    public static TestSuite suite()
    {
        return new TestSuite(ParserTest.class);
    }

    public ParserTest(String name)
    {
        super(name);
    }

    // ------------------------------------------------------------------------
    // test environment
    // ------------------------------------------------------------------------

    public void setUp()
    {
        parser = ParserFactory.INSTANCE.newJavaCCParser();
        sf = SyntaxFactory.INSTANCE;
    }

    private Parser parser;

    private SyntaxFactory sf;

    private SourceRef sr;

    private Node parse(String code)
    {
        try {
            return parser.parse(new StringSourceCode(code));
        }
        catch (ParserException ex) {
            fail("parsing failed: " + ex);
        }
        catch (Throwable ex) {
            fail("Unexpected exception: " + ex);
        }
        return null;
    }

    private void assertNonParsable(String code)
    {
        try {
            parser.parse(new StringSourceCode(code));
            fail("This source code should throw a ParseException");
        }
        catch (ParserException ex) {
        }
        catch (Throwable ex) {
            fail("Unexpected exception: " + ex);
        }
    }

    private void assertParse(Node expected, String code)
    {
        assertEquals(expected, parse(code));
    }

    private Node[] toNodeArray(ArrayList list)
    {
        return (Node[]) list.toArray(new Node[list.size()]);
    }

    private String[] toStringArray(ArrayList list)
    {
        return (String[]) list.toArray(new String[list.size()]);
    }

    // ------------------------------------------------------------------------
    // tests
    // ------------------------------------------------------------------------

    public void testAbstraction()
    {
        assertParse(
            sf.newAbstraction(sr, sf.newLabel(sr, "x"), ""),
            "\\ : x"
        );
        assertParse(
            sf.newAbstraction(sr, sf.newLabel(sr, "x"), "param"),
            "\\param : x"
        );
        assertParse(
            sf.newAbstraction(sr, sf.newLabel(sr, "x"), ""),
            "\\ (): x"
        );
        ArrayList params = new ArrayList();
        params.add("x");
        params.add("");
        assertParse(
            sf.newCurriedAbstraction(
                sr, sf.newLabel(sr, "x"), toStringArray(params)
            ),
            "\\x () : x"
        );
        params.add("b");
        assertParse(
            sf.newCurriedAbstraction(
                sr, sf.newLabel(sr, "x"), toStringArray(params)
            ),
            "\\ x () b: x"
        );
    }

    public void testApplication()
    {
        assertParse(sf.newApplication(
            sr, sf.newLabel(sr, "functor"), sf.newLabel(sr, "arg")),
            "functor arg"
        );
    }

    public void testBinding()
    {
        assertParse(
            sf.newBinding(sr, sf.newEmpty(sr), "a"),
            "a = ()"
        );
        assertParse(
            sf.newExtension(sr,
            sf.newBinding(sr, sf.newEmpty(sr), "a"), sf.newLabel(sr, "b")),
            "(a = ()), b"
        );
        ArrayList labels = new ArrayList();
        labels.add("f");
        assertParse(
            sf.newCompoundBinding(
                sr, sf.newLabel(sr, "x"), sf.newLabel(sr, "f"),
                toStringArray(labels), true
            ),
            "def f = x, f"
        );
        labels.add("g");
        assertParse(
            sf.newCompoundBinding(
                sr, sf.newLabel(sr, "x"), sf.newEmpty(sr),
                toStringArray(labels), false
            ),
            "f.g = x"
        );
        assertParse(
            sf.newCompoundBinding(
                sr, sf.newLabel(sr, "x"), sf.newEmpty(sr),
                toStringArray(labels), true
            ),
            "def f.g = x"
        );
        assertParse(
            sf.newFixedPoint(sr, sf.newEmpty(sr), "a"),
            "def a = ()"
        );
        labels.clear();
        labels.add("root");
        labels.add("label");
        assertParse(
            sf.newCompoundBinding(
                sr, sf.newEmpty(sr), sf.newEmpty(sr),
                toStringArray(labels), false
            ),
            "root.label = ()"
        );
        labels.clear();
        labels.add("dynamic");
        labels.add("label");
        assertParse(
            sf.newCompoundBinding(
                sr, sf.newEmpty(sr), sf.newEmpty(sr),
                toStringArray(labels), false
            ),
            "dynamic.label = ()"
        );
    }

    public void testCollection()
    {
        ArrayList children = new ArrayList();
        assertParse(
            sf.newCollection(sr, toNodeArray(children), "[", "]"),
            "[]"
        );
        children.add(sf.newBinding(sr, sf.newLabel(sr, "b"), "a"));
        assertParse(
            sf.newCollection(sr, toNodeArray(children), "[", "]"),
            "[a=b]"
        );
        children.add(sf.newLabel(sr, "c"));
        assertParse(
            sf.newCollection(sr, toNodeArray(children), "[", "]"),
            "[a=b, c]"
        );
    }

    public void testConstantNumber()
    {
        assertParse(
            sf.newConstant(sr, new Integer(17)),
            "17"
        );
        assertParse(
            sf.newConstant(sr, new Float(3.141)),
            "3.141"
        );
    }

    public void testConstantString()
    {
        assertParse(sf.newConstant(sr, "hello"), "\"hello\"");
        // new line
        assertParse(sf.newConstant(sr, "hel\nlo"), "\"hel\\nlo\"");
        // tab
        assertParse(sf.newConstant(sr, "hel\tlo"), "\"hel\\tlo\"");
        // backslash
        assertParse(sf.newConstant(sr, "hel\\lo"), "\"hel\\\\lo\"");
        // double quote
        assertParse(sf.newConstant(sr, "hel\"lo"), "\"hel\\\"lo\"");
        // octal char code
        assertParse(sf.newConstant(sr, "hel\077lo"), "\"hel\\077lo\"");
        // escaped new line (unix)
        assertParse(sf.newConstant(sr, "helloworld"), "\"hello\\\nworld\"");
        // escaped new line (doc)
        assertParse(sf.newConstant(sr, "helloworld"), "\"hello\\\n\rworld\"");
        // escaped new line (mac)
        assertParse(sf.newConstant(sr, "helloworld"), "\"hello\\\rworld\"");
        // invalid escape x
        assertParse(sf.newConstant(sr, "helloxworld"), "\"hello\\xworld\"");
        // invalid octal escape
        assertParse(sf.newConstant(sr, "hello17world"), "\"hello\\17world\"");
        // invalid octal escape at end of string
        assertParse(sf.newConstant(sr, "hello17"), "\"hello\\17\"");
    }

    public void testEmpty()
    {
        assertParse(sf.newEmpty(sr), "()");
    }

    public void testExtension()
    {
        assertParse(
            sf.newExtension(sr, sf.newEmpty(sr), sf.newEmpty(sr)),
            "(), ()"
        );

    }

    public void testInfixApplication()
    {
        assertParse(sf.newInfixApplication(
            sr, sf.newLabel(sr, "x"), sf.newLabel(sr, "y"), "+"),
            "x + y"
        );
        // operator before new line
        assertParse(sf.newInfixApplication(
            sr, sf.newLabel(sr, "x"), sf.newLabel(sr, "y"), "+"),
            "x+\ny"
        );
    }

    public void testLabel()
    {
        String helloLabel = "hello";
        String underscoreLabel = "hello_world";
        String infixLabel = "_+_";
        String prefixLabel = "+_";
        String newCollLabel = "[_]";
        String defaultInfixLabel = "_+_default";
        String defaultPrefixLabel = "+_default";

        assertParse(sf.newLabel(sr, helloLabel), helloLabel);
        assertParse(sf.newLabel(sr, underscoreLabel), underscoreLabel);
        assertParse(sf.newLabel(sr, infixLabel), infixLabel);
        assertParse(sf.newLabel(sr, prefixLabel), prefixLabel);
        assertParse(sf.newLabel(sr, newCollLabel), newCollLabel);
        assertParse(sf.newLabel(sr, defaultInfixLabel), defaultInfixLabel);
        assertParse(sf.newLabel(sr, defaultPrefixLabel), defaultPrefixLabel);
    }

    public void testNamedAbstraction()
    {
        ArrayList labels = new ArrayList();
        labels.add("test");
        ArrayList params = new ArrayList();
        params.add("");
        assertParse(
            sf.newNamedAbstraction(
                sr, sf.newEmpty(sr), sf.newEmpty(sr), toStringArray(labels),
                toStringArray(params), false
            ),
            "test: ()"
        );
    }

    public void testPrefixApplication()
    {
        assertParse(
            sf.newPrefixApplication(sr, sf.newLabel(sr, "x"), "+"),
            "+ x"
        );
        // operator after new line
        assertParse(
            sf.newExtension(sr, sf.newLabel(sr, "f"),
                sf.newPrefixApplication(sr, sf.newLabel(sr, "g"), "+")),
            "f\n+g"
        );
        // operator combined with projection
        assertParse(
            sf.newProjection(sr,
                sf.newPrefixApplication(sr, sf.newLabel(sr, "f"), "*"), "x"),
            "*f.x"
        );
    }

    public void testProjection()
    {
        assertParse(
            sf.newProjection(sr, sf.newLabel(sr, "a"), "b"),
            "a.b"
        );
    }

    public void testQuote()
    {
        assertParse(
            sf.newQuote(
                sr, sf.newBinding(sr, sf.newEmpty(sr), "f"),
                sf.newLabel(sr, "f")
            ),
            "'f = (), f"
        );
        assertParse(
            sf.newQuote(
                sr, sf.newQuote(sr, sf.newLabel(sr, "func"), sf.newEmpty(sr)),
                sf.newLabel(sr, "x")
            ),
            "''func, x"
        );
    }

    public void testRoot()
    {
        assertParse(sf.newRoot(sr), "root");
    }

    public void testSandbox()
    {
        assertParse(
            sf.newSandbox(sr, sf.newEmpty(sr), sf.newLabel(sr, "x")),
            "root = (), x"
        );
    }

    public void testNonParsable()
    {
        // comma cannot be used as operator
        assertNonParsable("_,_");
        // root cannot be used as identifier
        assertNonParsable("hello.root");
        assertNonParsable("func root: ()");
    }

    public void testPrecedenceApplication()
    {
        assertParse(
            sf.newApplication(sr, sf.newLabel(sr, "p"),
                sf.newInfixApplication(
                    sr, sf.newLabel(sr, "a"), sf.newLabel(sr, "b"), "+"
                )
            ),
            "p a + b"
        );
    }

    public void testPrecedenceInfix()
    {
        assertParse(
            sf.newInfixApplication(sr, sf.newLabel(sr, "p"),
                sf.newInfixApplication(
                    sr, sf.newLabel(sr, "a"), sf.newLabel(sr, "b"), "+"
                ),
            "|"),
            "p | a + b"
        );
        assertParse(
            sf.newInfixApplication(sr,
                sf.newInfixApplication(
                    sr, sf.newLabel(sr, "a"), sf.newLabel(sr, "b"), "+"
                ),
            sf.newLabel(sr, "p"), "|"),
            "a + b | p"
        );
        assertParse(
            sf.newInfixApplication(sr,
                sf.newInfixApplication(
                    sr, sf.newLabel(sr, "f"), sf.newLabel(sr, "g"), "*"
                ),
                sf.newInfixApplication(
                    sr, sf.newLabel(sr, "h"), sf.newLabel(sr, "i"), "*"
                ),
            "+"),
            "f * g + h * i"
        );
        assertParse(
            sf.newInfixApplication(sr,
                sf.newInfixApplication(
                    sr, sf.newLabel(sr, "f"), sf.newLabel(sr, "g"), "*"
                ),
                sf.newInfixApplication(
                    sr,
                    sf.newPrefixApplication(sr, sf.newLabel(sr, "h"), "+"),
                    sf.newLabel(sr, "i"), "*"
                ),
            "+"),
            "f * g + +h * i"
        );
    }

    // ------------------------------------------------------------------------

/*

    public void testNestedAssignment2() {
        Node f = parse("a, f.g.y = x");
        Node exp = parse("a, f = (f, g = (f.g, y = x))");
        assertEquals(exp, f);
    }

    public void testNestedInRoot1() {
        Node f = parse("a, root.y = x");
        Node exp = parse("a, root = (root, y = x)");
        assertEquals(exp, f);
    }



   public void testNestedScope() {
        Node f = parse("a, def f.y = x");
        Node exp = parse("a, f = (f, def y = x)");
        assertEquals(exp, f);
    }

   public void testNestedFunctions() {
        Node f = parse("a, def f.y(): x");
        Node exp = parse("a, f = (f, def y(): x)");
        assertEquals(exp, f);
    }


    public void testNestedAbstraction() {
        Node f = parse("a, f.y(): x");
        Node exp = parse("a, f = (f, y(): x)");
        assertEquals(exp, f);
    }


    public void testNestedQuotedAbstraction() {
        Node f = parse("'DefaultOp.a(): b");
        Node exp = parse("root = (root, DefaultOp = (DefaultOp, a(): b))");
        assertEquals(exp, f);
    }



    }


    public void testGotcha3() {
        Node f = parse("p a.b");
        Node exp = parse("p(a.b)");
        assertEquals(exp, f);
    }

    public void testGotcha4() {
        Node f = parse("a().b().c()");
        Node exp = parse("((((a()).b)()).c)()");
        assertEquals(exp, f);
    }

    public void testGotcha5() {
        Node f = parse("f a (b)");
        Node exp = parse("(f a)b");
        assertEquals(exp, f);
    }

    public void testGotcha6() {
        Node f = parse("f a(b)");
        Node exp = parse("f(a b)");
        assertEquals(exp, f);
    }

    public void testGotcha7() {
        Node f = parse("f a b");
        Node exp = parse("(f a)b");
        assertEquals(exp, f);
    }

    public void testGotcha8() {
        Node f = parse("f()+b");
        Node exp = parse("(f()) + b");
        assertEquals(exp, f);
    }

    public void testGotcha9() {
        Node f = parse("f ()+b");
        Node exp = parse("f(() + b)");
        assertEquals(exp, f);
    }

    public void testPrefixInfix1() {
        Node f = parse("f + | g");
        Node exp = parse("f + (| g)");
        assertEquals(exp, f);
    }


    public void testApplicationProjection() {
        Node f = parse("f a.b");
        Node exp = parse("f (a.b)");
        assertEquals(exp, f);
    }


    public void testPrecedencePrefix() {
        Node f = parse("+f()");
        Node exp = parse("(+f)()");
        assertEquals(exp, f);
    }



*/

}