package ch.unibe.piccola.syntax;

import java.util.Arrays;
import ch.unibe.piccola.source.SourceRef;

/**
 * Represents a collection syntax node.
 *
 * @author   Stefan Kneubuehl
 * @version  3.4c (Nov 14, 2002)
**/
public class Collection extends Node
{

    // ------------------------------------------------------------------------
    // public
    // ------------------------------------------------------------------------

    /**
     * Accepts a node visitor.
     *
     * @param  visitor the node visitor to be accepted.
     *
     * @return the result of the visitor's visitCollection method.
    **/
    public Object accept(NodeVisitor visitor)
    {
        return visitor.visitCollection(this);
    }

    /**
     * Compares this collection node with the specified object.
     *
     * @param  object the object to be compared with this node.
     *
     * @return <code>true</code> if this collection node is equal to object,
     *         <code>false</code> otherwise.
    **/
    public boolean equals(Object object)
    {
        if (!super.equals(object)) {
            return false;
        }
        Collection other = (Collection) object;
        return m_opening.equals(other.m_opening) &&
               m_closing.equals(other.m_closing) &&
               Arrays.equals(m_children, other.m_children);
    }

    /**
     * Returns the children of this collection node.
     *
     * @return the children of this collection node.
    **/
    public Node[] getChildren()
    {
        return m_children;
    }

    /**
     * Returns the closing braces/brackets combination of this collection node.
     *
     * @return the closing braces/brackets combination of this collection node.
    **/
    public String getClosing()
    {
        return m_closing;
    }

    /**
     * Returns the constructor label for this collection node.
     *
     * @return the constructor label for this collection node.
    **/
    public String getLabel()
    {
        return m_opening + "_" + m_closing;
    }

    /**
     * Returns the opening braces/brackets combination of this collection node.
     *
     * @return the opening braces/brackets combination of this collection node.
    **/
    public String getOpening()
    {
        return m_opening;
    }

    /**
     * Returns a debug string representation of this collection node.
     *
     * @return a debug string representation of this collection node.
    **/
    public String toString()
    {
        return "Collection(" + childrenString() + ")";
    }

    public Node transform(NodeVisitor visitor)
    {
        Node[] newChildren = new Node[m_children.length];
        for (int i = 0; i < m_children.length; ++i) {
            newChildren[i] = m_children[i].transform(visitor);
        }
        return new Collection(
            getSourceRef(), newChildren, m_opening, m_closing
        );
    }

    // ------------------------------------------------------------------------
    // default
    // ------------------------------------------------------------------------

    Collection(SourceRef sourceRef, Node[] children,
               String opening, String closing)
    {
        super(sourceRef);
        m_opening = opening;
        m_children = children;
        m_closing = closing;
    }

    // ------------------------------------------------------------------------
    // private
    // ------------------------------------------------------------------------

    private final Node[] m_children;

    private final String m_opening;

    private final String m_closing;

    private final String childrenString()
    {
        StringBuffer result = new StringBuffer();
        if (0 < m_children.length) {
            result.append(m_children[0]);
        }
        for (int i = 1; i < m_children.length; ++i) {
            result.append(", " + m_children[i]);
        }
        return result.toString();
    }

}