package ch.unibe.piccola.type;

/**
 * This class represents a contractual type 'P{R}'.
 *
 * @author    Stefan Kneubuehl
 * @version   3.4c (Jan 04, 2003)
**/
public class ContractualType
{

    // ------------------------------------------------------------------------
    // public static
    // ------------------------------------------------------------------------

    /**
     * Returns a new contractual type representing the empty type '(){()}'.
    **/
    public static final ContractualType newEmpty()
    {
        return new ContractualType(Type.EMPTY);
    }

    /**
     * Returns a new contractual type representing the root type 'R{R}'.
    **/
    public static final ContractualType newRoot()
    {
        return new ContractualType(Type.ROOT);
    }

    /**
     * Returns a new contractual type representing the type 'a{x:a}'.
    **/
    public static ContractualType newLookup(String x)
    {
        TypeVariable var = Type.newVariable();
        Constraints cs = new Constraints();
        cs.add(Type.ROOT, var.reverse().bind(x));
        return new ContractualType(var, cs);
    }

    // ------------------------------------------------------------------------
    // public
    // ------------------------------------------------------------------------

    public ContractualType(Type provided)
    {
        this(provided, new Constraints());
    }

    /**
     * Implements inference rule 'Apply'
    **/
    public ContractualType apply(ContractualType type)
    {
        debugMsg("Apply");
        TypeVariable var = Type.newVariable();
        Constraints cs = new Constraints();
        cs.addAll(m_constraints);
        cs.addAll(type.m_constraints);
        cs.add(m_provided, type.m_provided.service(var.reverse()));
        return new ContractualType(var, cs);
    }

    /**
     * Implements inference rule 'Close'
    **/
    public ContractualType close(ContractualType type)
    {
        debugMsg("Close");
        Constraints cs = new Constraints();
        cs.addAll(type.m_constraints);
        cs.subst(Type.ROOT, m_provided);
        cs.addAll(m_constraints);
        return new ContractualType(type.m_provided.subst(Type.ROOT, m_provided), cs);
    }

    /**
     * Implements inference rule 'Extend'
    **/
    public ContractualType extend(ContractualType type)
    {
        debugMsg("Extend");
        Constraints cs = new Constraints();
        cs.addAll(m_constraints);
        cs.addAll(type.m_constraints);
        return new ContractualType(m_provided.extend(type.m_provided), cs);
    }

    /**
     * Implements inference rule 'Restrict'
    **/
    public ContractualType restrict(String x)
    {
        debugMsg("Restrict");
        Constraints cs = new Constraints();
        cs.addAll(m_constraints);
        return new ContractualType(m_provided.hide(x), cs);
    }

    /**
     * Implements inference rule 'Label'
    **/
    public ContractualType label(String x)
    {
        debugMsg("Label");
        Constraints cs = new Constraints();
        cs.addAll(m_constraints);
        return new ContractualType(m_provided.bind(x), cs);
    }

    public Type getProvided()
    {
        return m_provided;
    }

    /**
     * Implements inference rule 'Service'
    **/
    public ContractualType service(String x)
    {
        debugMsg("Abstract");
        Constraints cs = new Constraints();
        cs.addAll(m_constraints);
        TypeVariable var = Type.newVariable();
        Type replace = Type.ROOT.extend(var.bind(x));
        cs.subst(Type.ROOT, replace);
        cs.add(var, Type.EMPTY);
        replace = Type.ROOT.extend(var.reverse().bind(x));
        return new ContractualType(var.reverse().service(m_provided.subst(Type.ROOT, replace)), cs);
//        cs.opRemoveLabel(x);
//        return new ContractualType(m_constraints.opLookupLabel(x).service(m_provided), cs);
    }

    public void subst(TypeVariable var, Type replace)
    {
        debugMsg(toString() + "[" + var + " -> " + replace + "]");
        m_provided = m_provided.subst(var, replace);
        m_constraints.subst(var, replace);
    }

    public String toString()
    {
        return m_provided.toString() +
            (m_constraints.isEmpty() ? "" : " | ") + m_constraints.toString();
    }

    /**
     * Implements the type-checking algorithm.
    **/
    public void typeCheck()
    {
        int i = 1;
        debugMsg("type checking: " + toString());
        m_constraints.normalize();
        while (m_constraints.isResolvable(this)) {
            m_constraints.normalize();
            m_constraints.resolve(this);
            m_constraints.normalize();
            debugMsg("After " + i + " pass(es): " + toString());
            ++i;
        }
    }

    // ------------------------------------------------------------------------
    // private
    // ------------------------------------------------------------------------

    public ContractualType(Type provided, Constraints constraints)
    {
        m_provided = provided;
        m_constraints = constraints;
        typeCheck();
        debugMsg(this.toString());
    }

    private Type m_provided;

    private final Constraints m_constraints;

    private void debugMsg(String msg)
    {
//        System.out.println("    " + msg);
    }

}