package ch.unibe.piccola.vm;

import java.util.ArrayList;
import ch.unibe.piccola.code.Arguments;
import ch.unibe.piccola.code.Code;
import ch.unibe.piccola.code.CodeFrame;
import ch.unibe.piccola.code.IVM;
import ch.unibe.piccola.code.OpCode;
import ch.unibe.piccola.code.JavaOp;
import ch.unibe.piccola.source.SourceRef;
import ch.unibe.piccola.syntax.Syntax;
import ch.unibe.piccola.bridge.PiccolaList;

/**
 * @author   Stefan Kneubuehl
 * @version  3.6c (May 11, 2003)
**/
public class VM extends Agent implements IVM
{

    // ------------------------------------------------------------------------
    // private static
    // ------------------------------------------------------------------------

    private static final String LABEL_ARG = Syntax.labelArg();

    private static final String LABEL_CALLER = Syntax.labelCaller();

    private static final String LABEL_DYNAMIC = Syntax.labelDynamic();

    private static final String LABEL_SRC = Syntax.labelSrc();

    // ------------------------------------------------------------------------
    // public (opcodes)
    // ------------------------------------------------------------------------

    public final void opApply()
    {
        Form dynamic = popValue();
        Form service = popValue();
        Form caller = dynamic.projectDef(LABEL_CALLER, m_pool.m_factory.emptyForm()).bind(LABEL_CALLER);
        caller = caller.extend(m_pool.up(getSourceRef()).bind(LABEL_SRC));
        dynamic = dynamic.extend(caller.bind(LABEL_CALLER));
        pushCode(service.service().getCode());
        pushContext(service.service().getContext().extend(dynamic.bind(LABEL_DYNAMIC)));
    }

    public final void opBind(String label)
    {
        pushValue(popValue().bind(label));
    }

    public final void opEmpty()
    {
        pushValue(m_pool.m_factory.emptyForm());
    }

    public final void opExit()
    {
        popCode();
        popContext();
        if (sizeCode() <= m_wrapping) {
            m_wrapping = -1;
        }
    }

    public final void opExtend()
    {
        Form right = popValue();
        Form left = popValue();
        pushValue(left.extend(right));
    }

    public final void opHide(String label)
    {
        pushValue(popValue().hide(label));
    }

    public final void opJavaCall(JavaOp javaOp, Object target)
    {
        Form argument = getDynamic().projectDef(LABEL_ARG, m_pool.m_factory.emptyForm());
        pushValue(m_pool.up(javaOp.call(target, downArguments(argument))));
    }

    public final void opJavaObject(Object object)
    {
        pushValue(m_pool.up(object));
    }

    public final void opLookup(String label)
    {
        pushValue(getContext().project(label));
    }

    public final void opMeta()
    {
        pushValue(m_pool.wrap(popValue()));
    }

    public final void opPopContext()
    {
        popContext();
    }

    public final void opPopValue()
    {
        popValue();
    }

    public final void opProject(String label)
    {
        pushValue(popValue().project(label));
    }

    public final void opPushContext()
    {
        pushContext(popValue());
    }

    public final void opReceive(Object channel)
    {
        if (!(channel instanceof ChannelForm)) {
            throw new AgentException("opReceive: Channel expected.");
        }
        ((ChannelForm) channel).listen(this);
    }

    public final void opRoot()
    {
        pushValue(getContext());
    }

    public final void opSend(Object channel)
    {
        if (!(channel instanceof ChannelForm)) {
            throw new AgentException("opSend: Channel expected.");
        }
        Form form = popValue();
        ((ChannelForm) channel).send(form);
        opEmpty();
    }

    public final void opService(Code code)
    {
        pushValue(m_pool.newService(code, getContext()));
    }

    public final void opVariable()
    {
        pushValue(m_pool.newVariable(popValue()));
    }

    // ------------------------------------------------------------------------
    // default
    // ------------------------------------------------------------------------

    VM(VMPool pool, Code code, Form context, Form dynamic)
    {
        super(pool);
        m_exception = null;
        m_pool = pool;
        m_wrapping = -1;
        pushCode(code);
        context = context.extend(dynamic.bind(LABEL_DYNAMIC));
        pushContext(context);
        pushValue(m_pool.m_factory.emptyForm());
    }

    void pushValue(Form value)
    {
        value.push(this);
    }

    void doPushForm(Form value)
    {
        super.pushValue(value);
    }

    void doPushPeerForm(Form form, Class cl)
    {
        Form wrapper = m_pool.getPiccolaWrapper(cl);
        if (-1 == m_wrapping && null != wrapper) {
            m_wrapping = sizeCode();
            doPushForm(wrapper);
            opLookup(LABEL_DYNAMIC);
            doPushForm(form);
            opBind(LABEL_ARG);
            opExtend();
            opApply();
        }
        else {
            doPushForm(form);
        }
    }

    // ------------------------------------------------------------------------
    // protected
    // ------------------------------------------------------------------------

    protected void handleException(AgentException ex)
    {
        try {
            Thread.sleep(500);
        }
        catch (InterruptedException x) {}
        ex.setSourceRef(getSourceRef());
        ex.setDynamic(getDynamic());
        Form handler = m_pool.getExceptionHandler();
        if (null != m_exception || null == handler) {
            m_pool.println("Original exception: ");
            super.handleException(m_exception);
            m_pool.println("Exception while handling exception:");

            super.handleException(ex);
            SourceRef[] trace = stackTrace();
            for (int i = 0; i < trace.length; ++i) {
                if (null != trace[i]) m_pool.println(trace[i].toString());
            }

        }
        else {
            try {
                m_exception = ex;
                doPushForm(handler);
                opLookup(LABEL_DYNAMIC);
                opJavaObject(ex);
                opBind(LABEL_ARG);
                opExtend();
                opApply();
            }
            catch (AgentException x) {
                handleException(x);
            }
        }
    }

    // ------------------------------------------------------------------------
    // private
    // ------------------------------------------------------------------------

    private AgentException m_exception;

    private int m_wrapping;

    private final VMPool m_pool;

    private final Arguments downArguments(Form argument)
    {
        Arguments args = new Arguments();
        Object downArg = argument.down();
        if (downArg instanceof PiccolaList) {
            PiccolaList list = (PiccolaList) downArg;
            for (int i = 0; i < list.size(); ++i) {
                args.add(((Form)list.get(i)).down());
            }
            return args;
        }
        int i = 1;
        Form arg = argument.projectDef(Syntax.labelVal(i), null);
        while (null != arg) {
            args.add(arg.down());
            ++i;
            arg = argument.projectDef(Syntax.labelVal(i), null);
        }
        if (args.size() == 0 && !argument.isEmpty()) {
            args.add(argument.down());
        }
        return args;
    }

}

